function _inheritsLoose(subClass, superClass) { subClass.prototype = Object.create(superClass.prototype); subClass.prototype.constructor = subClass; _setPrototypeOf(subClass, superClass); }
function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }
import View from './View.js';

// This options keys will be extended on view instance.
var componentOptions = {
  key: true,
  state: true,
  onCreate: true,
  onChange: true,
  onRender: true
};

/*
 * Helper function. Extract attributes from html tag
 * @param text {string} html text
 * @return {object} Object with keys / values representing attributes.
 */
var extractAttributes = function extractAttributes(text) {
  var attributes = {};
  var re = /([\w|data-]+)(?:=["']?((?:.(?!["']?\s+(?:\S+)=|\s*\/?[>"']))+.)["']?)?/g;
  var result;
  while ((result = re.exec(text)) !== null) {
    attributes[result[1]] = typeof result[2] === 'undefined' ? true : result[2];
  }
  return attributes;
};

/*
 * Helper function. If first arg is a placeholder for an expression, return the expression.
 */
var getExpression = function getExpression(placeholder, expressions) {
  var match = placeholder && placeholder.match && placeholder.match(new RegExp(Component.EXPRESSION_PLACEHOLDER_TEMPLATE('(\\d+)')));
  return match && match[1] ? expressions[match[1]] : placeholder;
};

/*
 * Helper function. If expression is a function, call it with context and args.
 */
var evalExpression = function evalExpression(expression, context) {
  for (var _len = arguments.length, args = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
    args[_key - 2] = arguments[_key];
  }
  return typeof expression === 'function' ? expression.apply(context, args) : expression;
};

/**
 * Components are a special kind of `View` that is designed to be easily composable, 
 * making it simple to add child views and build complex user interfaces.  
 * Unlike views, which are render-agnostic, components have a specific set of rendering 
 * guidelines that allow for a more declarative development style.  
 * Components are defined with the `create` static method, which takes a tagged template.
 * @module
 * @extends Rasti.View
 * @param {object} options Object containing options. The following keys will be merged to `this`: model, state, key, onDestroy, onRender, onCreate, onChange.
 * @property {string} key A unique key to identify the component. Used to recycle child components.
 * @property {object} model A `Rasti.Model` or any emitter object containing data and business logic.
 * @property {object} state A `Rasti.Model` or any emitter object containing data and business logic, to be used as internal state.
 * @example
 * import { Component, Model } from 'rasti';
 * // Create Timer component.
 * const Timer = Component.create`
 *     <div>
 *         Seconds: <span>${({ model }) => model.seconds}</span>
 *     </div>
 * `;
 * // Create model to store seconds.
 * const model = new Model({ seconds: 0 });
 * // Mount timer on body.
 * Timer.mount({ model }, document.body);
 * // Increment `model.seconds` every second.
 * setInterval(() => model.seconds++, 1000);
 */
var Component = /*#__PURE__*/function (_View) {
  function Component() {
    var _this;
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    _this = _View.apply(this, arguments) || this;
    // Extend "this" with options, mapping componentOptions keys.
    Object.keys(options).forEach(function (key) {
      if (componentOptions[key]) _this[key] = options[key];
    });
    // Store options by default.
    _this.options = options;
    // Bind onChange to this to be used as listener.
    // Store bound version, so it can be removed on onDestroy method.
    _this.onChange = _this.onChange.bind(_this);
    // Listen to model changes and call onChange.
    if (_this.model && _this.model.on) _this.model.on('change', _this.onChange);
    if (_this.state && _this.state.on) _this.state.on('change', _this.onChange);
    // Call life cycle method.
    _this.onCreate.apply(_this, arguments);
    return _this;
  }

  /*
   * Override. We don't want to ensure an element on instantiation.
   * We will provide it later.
   */
  _inheritsLoose(Component, _View);
  var _proto = Component.prototype;
  _proto.ensureElement = function ensureElement() {
    // If el is provided, delegate events.
    if (this.el) {
      this.delegateEvents();
      this.id = this.el.id;
    }
    // Ensure id.
    if (!this.id) {
      this.id = this.attributes && this.attributes.id ?
      // If id is provided, evaluate it.
      evalExpression(this.attributes.id, this, this) :
      // Generate a unique id and set it as id attribute.
      Component.ID_TEMPLATE(this.uid);
    }
  }

  /*
   * Find view's element on parent node, using id.
   */;
  _proto.findElement = function findElement(parent) {
    return (parent || document).querySelector("#" + this.id);
  }

  /*
   * Eval attributes expressions.
   */;
  _proto.getAttributes = function getAttributes() {
    var _this2 = this;
    var add = {
      id: this.id
    };
    var remove = {};
    var html = ["id=\"" + this.id + "\""];
    if (this.attributes) {
      Object.keys(this.attributes).forEach(function (key) {
        if (key === 'id') return;
        // Evaluate attribute value.
        var value = evalExpression(_this2.attributes[key], _this2, _this2);

        // Transform bool attribute values
        if (value === false) {
          remove[key] = true;
        } else if (value === true) {
          add[key] = '';
          html.push(key);
        } else {
          if (value === null || typeof value === 'undefined') value = '';
          add[key] = value;
          html.push(key + "=\"" + value + "\"");
        }
      });
    }
    return {
      add: add,
      remove: remove,
      html: html.join(' ')
    };
  }

  /*
   * Used internally on the render process.
   * Attach the view to the dom element.
   */;
  _proto.hydrate = function hydrate(parent) {
    var _this3 = this;
    this.el = this.findElement(parent);
    this.delegateEvents();
    this.children.forEach(function (child) {
      return child.hydrate(_this3.el);
    });
    this.onRender.call(this, 'hydrate');
  }

  /*
   * Used internally on the render process.
   * Reuse a view that has `key` when its parent is rendered.
   */;
  _proto.recycle = function recycle(parent) {
    // Find element to be replaced. It has same id.
    var toBeReplaced = this.findElement(parent);
    // Replace it with this.el.
    toBeReplaced.replaceWith(this.el);
    // Call `onRender` lifecycle method.
    this.onRender.call(this, 'recycle');
  }

  /*
   * Override. Add some custom logic to super `destroy` method.
   */;
  _proto.destroy = function destroy() {
    _View.prototype.destroy.apply(this, arguments);
    // Stop listening to `change`.
    if (this.model && this.model.off) this.model.off('change', this.onChange);
    if (this.state && this.state.off) this.state.off('change', this.onChange);
    // Set destroyed flag to prevent a last render after destroyed.
    this.destroyed = true;
    // Return `this` for chaining.
    return this;
  }

  /**
   * Lifecycle method. Called when the view is created at the end of the constructor.
   * @param options {object} The view options.
   */;
  _proto.onCreate = function onCreate() {}

  /**
   * Lifecycle method. Called when model emits `change` event.
   * By default calls render method.
   * This method should be extended with custom logic.
   * Maybe comparing new attributes with previous ones and calling
   * render when needed. Or doing some dom transformation.
   * @param model {Rasti.Model} The model that emitted the event.
   * @param changed {object} Object containing keys and values that has changed.
   * @param [...args] {any} Any extra arguments passed to set method.
   */;
  _proto.onChange = function onChange() {
    this.render();
  }

  /**
   * Lifecycle method. Called when the view is rendered.
   * @param type {string} The render type. Can be `render`, `hydrate` or `recycle`.
   */;
  _proto.onRender = function onRender() {}

  /**
   * Lifecycle method. Called when the view is destroyed.
   * @param {object} options Options object or any arguments passed to `destroy` method.
   */;
  _proto.onDestroy = function onDestroy() {}

  /*
   * Replace expressions.
   */;
  _proto.replaceExpressions = function replaceExpressions(string, addChild) {
    var _this4 = this;
    return string.replace(new RegExp(Component.EXPRESSION_PLACEHOLDER_TEMPLATE('(\\d+)'), 'g'), function (match) {
      var expression = getExpression(match, _this4.template.expressions);
      // Eval expression. Pass view as argument.
      var result = evalExpression(expression, _this4, _this4);
      // Treat all expressions as arrays.
      var results = result instanceof Array ? result : [result];
      // Replace expression with the result of the evaluation.
      return results.reduce(function (out, result) {
        var parsed;
        // If result is true, replace it with a placeholder.
        if (result === true) parsed = Component.TRUE_PLACEHOLDER;
        // If result is false, replace it with a placeholder.
        else if (result === false) parsed = Component.FALSE_PLACEHOLDER;
        // Replace null or undefined with empty string.
        else if (result === null || typeof result === 'undefined') parsed = '';
        // If result is a view, call addChild callback.
        else if (result && typeof result.render === 'function') parsed = addChild(result);
        // Return expression itself.
        else parsed = result;
        // Concatenate expressions.
        return out + parsed;
      }, '');
    })
    // Replace `attribute="true"` with `attribute` and `attribute="false"` with empty string.
    .replace(new RegExp("([\\w|data-]+)=([\"'])?(" + Component.TRUE_PLACEHOLDER + "|" + Component.FALSE_PLACEHOLDER + ")\\2", 'g'), function (match, attribute, quote, placeholder) {
      return placeholder === Component.TRUE_PLACEHOLDER ? attribute : '';
    })
    // Replace rest of true or false expressions with empty string.
    .replace(new RegExp(Component.TRUE_PLACEHOLDER + "|" + Component.FALSE_PLACEHOLDER, 'g'), '');
  }

  /*
   * Treat the whole view as a HTML string.
   */;
  _proto.toString = function toString() {
    var _this5 = this;
    // Normally there won't be any children, but if there are, destroy them.
    this.destroyChildren();
    // Get tag name.
    var tag = this.tag || 'div';
    // Get attributes.
    var attributes = this.getAttributes().html;
    // Replace expressions of inner template.
    var close = this.template && this.template.inner;
    var inner = close && this.replaceExpressions(this.template.inner, function (component) {
      // Add child component.
      return _this5.addChild(component);
    });
    // Generate outer template.
    return close ? "<" + tag + " " + attributes + ">" + inner + "</" + tag + ">" : "<" + tag + " " + attributes + " />";
  }

  /*
   * View render method.
   */;
  _proto.render = function render() {
    var _this6 = this;
    // Prevent a last re render if view is already destroyed.
    if (this.destroyed) return this;
    // If `this.el` is not present, create a new `this.tag` element.
    if (!this.el) {
      this.el = this.createElement(this.tag);
      this.delegateEvents();
    }
    // Set `this.el` attributes.
    var attributes = this.getAttributes();
    // Remove attributes.
    Object.keys(attributes.remove).forEach(function (key) {
      _this6.el.removeAttribute(key);
    });
    // Add attributes.
    Object.keys(attributes.add).forEach(function (key) {
      _this6.el.setAttribute(key, attributes.add[key]);
    });
    // Check for `template.inner` to see if view has innerHTML.
    if (this.template && this.template.inner) {
      var previousChildren = this.children;
      this.children = [];
      var nextChildren = [];
      var recycledChildren = [];
      // Store active element.
      var activeElement = document.activeElement;
      // Replace expressions. Set html inside of `this.el`.
      this.el.innerHTML = this.replaceExpressions(this.template.inner, function (component) {
        var out = component;
        // Check if child already exists.
        var found = component.key && previousChildren.find(function (previousChild) {
          return previousChild.key === component.key;
        });
        if (found) {
          var tag = found.el.tagName.toLowerCase();
          var id = found.el.id;
          // If child already exists, replace it html by its root element.
          out = "<" + tag + " id=\"" + id + "\"></" + tag + ">";
          // Add child to recycled children.
          recycledChildren.push(found);
          // Destroy new child component. Use recycled one instead.
          component.destroy();
        } else {
          // Not found. Add new child component.
          nextChildren.push(component);
        }
        // Component html.
        return out;
      });
      // Add new children. Hydrate them.
      nextChildren.forEach(function (nextChild) {
        _this6.addChild(nextChild).hydrate(_this6.el);
      });
      // Replace children root elements with recycled components.
      recycledChildren.forEach(function (recycledChild) {
        _this6.addChild(recycledChild).recycle(_this6.el);
      });
      // Destroy unused children.
      previousChildren.forEach(function (previousChild) {
        var found = recycledChildren.indexOf(previousChild) > -1;
        if (!found) previousChild.destroy();
      });
      // Restore focus.
      if (this.el.contains(activeElement)) {
        activeElement.focus();
      }
    }
    // Call onRender lifecycle method.
    this.onRender.call(this, 'render');
    // Return this for chaining.
    return this;
  }

  /**
   * Helper method used to extend a `Component`, creating a subclass.
   * @static
   * @param {object} object Object containing methods to be added to the new `Component` subclass. Also can be a function that receives the parent prototype and returns an object.
   */;
  Component.extend = function extend(object) {
    var Current = this;
    var Extended = /*#__PURE__*/function (_Current) {
      function Extended() {
        return _Current.apply(this, arguments) || this;
      }
      _inheritsLoose(Extended, _Current);
      return Extended;
    }(Current);
    Object.assign(Extended.prototype, typeof object === 'function' ? object(Current.prototype) : object);
    return Extended;
  }

  /**
   * Mount the component into the dom.
   * It instantiate the Component view using options, 
   * appends its element into the DOM (if `el` is provided).
   * And returns the view instance.
   * @static
   * @param {object} options The view options.
   * @param {node} el Dom element to append the view element.
   * @param {boolean} hydrate If true, the view will use existing html.
   * @return {Rasti.Component}
   */;
  Component.mount = function mount(options, el, hydrate) {
    // Instantiate view.
    var view = new this(options);
    // If `el` is passed, mount component.
    if (el) {
      if (hydrate) {
        view.toString();
      } else {
        var fragment = document.createElement('template');
        // Add html text into element inner html.
        fragment.innerHTML = view;
        // Add to dom.
        el.appendChild(fragment.content);
      }
      view.hydrate(el);
    }
    // Return view instance.
    return view;
  }

  /**
   * Takes a tagged template containing an HTML string, 
   * and returns a new `Component` class.
   * - The template outer tag and attributes will be used to create the view's root element.
   * - Boolean attributes should be passed in the form of `attribute="${() => true}"`.
   * - Event handlers should be passed, at the root element, in the form of `onEventName=${{'selector' : listener }}`. Where `selector` is a css selector. The event will be delegated to the view's root element.
   * - The template inner HTML will be used as the view's template.
   * - Template interpolations that are functions will be evaluated on the render process. Receiving the view instance as argument. And being bound to it.
   * - If the function returns `null`, `undefined`, `false` or empty string, the interpolation won't render any content.
   * - If the function returns a component instance, it will be added as a child component.
   * - If the function returns an array, each item will be evaluated as above.
   * @static
   * @param {string} HTML template for the component.
   * @return {Rasti.Component}
   */;
  Component.create = function create(strings) {
    for (var _len2 = arguments.length, expressions = new Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
      expressions[_key2 - 1] = arguments[_key2];
    }
    var parts = [];
    // Replace functions and objects interpolations with `{number}`.
    // Where `number` is the index on expressions array.
    strings.forEach(function (string, i) {
      // Add string part.
      parts.push(string);
      // Add expression placeholder for later or expression eval.
      if (expressions[i]) {
        parts.push(typeof expressions[i] === 'function' || typeof expressions[i] === 'object' ? Component.EXPRESSION_PLACEHOLDER_TEMPLATE(i) : expressions[i]);
      }
    });
    // Create output text for main template.
    var main = parts.join('').trim().replace(/\n/g, '');
    // Extract outer tag, attributes and inner html.
    var result = main.match(/^<([a-z]+[1-6]?)(.*?)>(.*)<\/\1>$/) || main.match(/^<([a-z]+)(.*?)\/>$/);
    // Parse attributes from html text into an object.
    var attributes = extractAttributes(result[2]);
    // Events to be delegated.
    var events = {};
    // Filter events. To generate events object.
    // Generate attributes object, replace placeholders with expressions.
    attributes = Object.keys(attributes).reduce(function (out, key) {
      // Is Event?
      var matchKey = key.match(/on(([A-Z]{1}[a-z]+)+)/);
      // Is placeholder for function or object?
      // Get expression or value.
      var value = getExpression(attributes[key], expressions);
      // Is event handler. Add to events object.
      if (matchKey && matchKey[1]) {
        var eventType = matchKey[1].toLowerCase();
        Object.keys(value).forEach(function (selector) {
          return events["" + eventType + (selector === '&' ? '' : " " + selector)] = value[selector];
        });
        return out;
      }
      // Is attribute. Add to attributes object.
      out[key] = value;
      return out;
    }, {});
    var Current = this;
    // Create subclass for this component.
    return Current.extend({
      // Set events.
      events: events,
      // Set attributes.
      attributes: attributes,
      // Set template.
      template: {
        // Template for innerHTML of root element.
        inner: result[3],
        // Template expressions.
        expressions: expressions
      },
      // Set root element tag.
      tag: result[1]
    });
  };
  return Component;
}(View);
export { Component as default };
Component.ID_TEMPLATE = function (uid) {
  return "rasti-component-" + uid;
};
Component.EXPRESSION_PLACEHOLDER_TEMPLATE = function (idx) {
  return "__RASTI_EXPRESSION{" + idx + "}";
};
Component.TRUE_PLACEHOLDER = '__RASTI_TRUE';
Component.FALSE_PLACEHOLDER = '__RASTI_FALSE';